/*
 *  $Id: IO.java,v 1.3 1996/01/10 21:51:52 iansmith Exp $
 *  $Author: iansmith $
 *  $Log: IO.java,v $
 * Revision 1.3  1996/01/10  21:51:52  iansmith
 * The principal change in this version was the modification of the reading
 * of string literals. This is now tested and working... it is not
 * terrific from a performance point of view as it does some recursion
 * with big strings as parameters, but deal with it.
 *
 * Currently, I've turned off protocol debugging.
 *
 * Revision 1.2  1996/01/08  05:56:19  iansmith
 * Modified the read string routine to now be able to handle literal strings.
 * These are strings where it is prefixed by a {N} where n is the number of
 * bytes of the string. This allows 8bit data to come through, not to mention
 * CR/LF pairs. This is not yet tested.
 *
 * Modified roundTrip to throw InternalErr if we get an IOException. I'm still
 * not sure this is right.
 *
 *  Revision 1.1  1996/01/06 19:50:35  iansmith
 *  Initial revision
 *
 */

package GIT.GVU.IMAP;
import java.util.*;
import java.net.*;
import java.lang.*;
import java.io.*;

/* This class is only visible to things in this package */

class IO {
  /* this hashtable keeps track of the pending client messages and */
  /* makes sure server responses match these request ids */
  private Hashtable pendingRequests;
  /* this is our socket to the server */
  private Socket serverSocket;
  /* the I/O streams for this socket */
  private InputStream input;
  private OutputStream output;
  /* client output string counter */
  private int xmitCounter;
  /* debugging the protcol switch */
  private static final boolean debugProto=false;
  /* this will hold our data that we are buffering on the input */
  /* side of the world */
  StringBuffer inputBuff;
  /* 
   * construct an IMAP I/O object 
   */ 
  public IO(String IMAPHost) throws IOException {
    /* initialize the hashtable */
    pendingRequests=new Hashtable(5);
    /* set up a server connection ... IMAP is always port 143 and TCP */
    serverSocket=new Socket(IMAPHost,143,true);
    /* initialize the input and output streams */
    input=serverSocket.getInputStream();
    output=serverSocket.getOutputStream();
    /* initialize the client counter */
    xmitCounter=0;
    /* set up an input buffer */
    inputBuff=new StringBuffer();
  }
  /*
   * This function lets you write a client command to the server 
   */
  public void writeString(String outValue) {
    /* just call the internal version and don't archive the result  */
    writeStringInternal(outValue,false); 
  }
  /*
   * This function lets you write a client command to the server 
   */
  void writeStringInternal(String outValue,boolean archive) {
    /* allocate the char array */
    byte theByte[];
    /* the string to be sent */
    String finalString;
    /* need a tokenizer to rip out first word */
    StringTokenizer tokenizer=new StringTokenizer(outValue," ",false);
    /* compute a string with the prefix on it */
    String prefix=new String("A" + xmitCounter + " ");
    finalString=new String(prefix+outValue);
    theByte=new byte[finalString.length() +2];
    /* do we need to archive it ? */
    if (archive==true) {
      /* archive in a hashtable so we can check later */
      /* XXX RIGHT NOW I DON'T USE THE VALUES XXX */
      pendingRequests.put(prefix.substring(0,prefix.length()-1), 
                          tokenizer.nextToken());
    }
    /*increment the prefix counter */
    xmitCounter+=1;
    /* convert to chars */
    finalString.getBytes(0,finalString.length(),theByte,0);
    /* add CR/LF */
    theByte[finalString.length()]='\r';
    theByte[finalString.length()+1]='\n';
    /* echo the protocol to the terminal */
    if (debugProto==true) {
      System.out.println("IMAP: --> " + finalString);
    }
    /* send it to the server */
    try {
      output.write(theByte);
      output.flush();
    } catch (IOException e){ 
      System.out.println("IO EXCEPTION!");
      return;
    }
  }
  /* reads a response from the server */
  String readString() throws InternalErr, IOException {
    /* this will tokenize the string for us */
    StringTokenizer st;
    /* we'll read up to 256 characters if we can */
    byte block[]= new byte[256];
    /* we'll use this to store how much we read */
    int actuallyRead=0;
    /* this controls whether or not we have found an cr/lf and can stop */
    boolean done;
    /* where the results go */
    char result[];
    /* where the final result goes */
    String finalString;
    /* temp copy of the input buffer */
    String tmpBuff;
    /* string to hold the number of a literal string */
    String num;
    /* extra part of a string which we add to form a literal string */
    String extra;
    /* number that we got from the server in a literal string */
    int n;
    /* number of bytes we've read in a literal string */
    int haveRead=0;
    /* where the left curly is in a literal string */
    int lcurly;

    /* initialize the done condition... we are already done */
    /* if the input buffer is already holding a CR/LF pair */
    tmpBuff=inputBuff.toString();
    if (tmpBuff.indexOf("\r\n")!=-1) {
       /* already have a line */
      done=true;
    } else {
      /* we need to read some chars */      
      done=false; 
    }
    /* loop until we have a string terminated with CR/LF */
    while (done==false) {
      /* try to read 256 bytes */
      actuallyRead=input.read(block,0,256);
      /* append to the buffer */
      inputBuff.append(new String(block,0,0,actuallyRead));
      /* did we get a line yet? */
      tmpBuff=inputBuff.toString();
      if (tmpBuff.indexOf("\r\n")!=-1) {
	done=true;
      }
    }

    /* we got a line, calculate part to return and shove it in result */
    result=new char[tmpBuff.indexOf("\r\n")];
    inputBuff.getChars(0,tmpBuff.indexOf("\r\n"),result,0);
    /* get rid of what we just returned */
    inputBuff=new StringBuffer(tmpBuff.substring(tmpBuff.indexOf("\r\n")+2,
                                                 tmpBuff.length()));
    /* set up for return */
    finalString=new String(result);
    /* ok, it could be the case that we've gotten a literal string... this will be */
    /* demarcated by a right curly brace at the last character position... */
    if ((finalString.length()>1) && 
        (finalString.charAt(finalString.length()-1)=='}')) {
      /* ok, we have a literal string ... work backward looking for the matching*/
      /* left curly brace */
      lcurly=finalString.lastIndexOf('{');
      /* did we find it ... if we didn't find it we are going to assume that this */
      /* really wasn't a literal string... I'm not sure this isn't a protocol error */
      /* this may be XXX */
      if (lcurly!=-1) {
	/* ok, found the left curly, rip out the number part of the string */
	num=finalString.substring(lcurly+1,finalString.length()-1);
	/* try to extract the number from it */
	try {
	  n=Integer.valueOf(num,10).intValue();
	} catch (NumberFormatException e) {
	  throw new InternalErr("Unable to find a number in a literal string from the server!");
	}
	/* possible empty string */
	if (n==0) {
	  /* don't bother doing any work*/
	  extra="";
	} else {
	  /* allocate a buffer to hold it */
	  block=new byte[n];
	  /* we might have enough bytes already in the buffer or we might just */
	  /* have part of it ... note: this affects how much we tried to read from */
	  /* the network below */
	  if (n<inputBuff.length()) {
	    haveRead=n;
	  } else {
	    haveRead=inputBuff.length();
	  }
	  /* convert the string buffer to a more usable plain string */
	  tmpBuff=inputBuff.toString();
	  /* grab our chars out */
	  tmpBuff.getBytes(0,haveRead,block,0);
	  /* clear the input buffer if we have used it up*/
	  if (haveRead==inputBuff.length()) {
	    inputBuff=new StringBuffer("");
	  } else {
	    /* just clear out the part we used */
	    inputBuff=new StringBuffer(tmpBuff.substring(haveRead));
	  }
	}
	/* loop until we read it all */
	while (haveRead<n) {
	  /* read in some more bytes */
	  actuallyRead=input.read(block,haveRead,n-haveRead);
	  /* sanity check */
	  if (actuallyRead==-1) {
	    throw new java.io.IOException("End of file reached on a read of the network to IMAP Server!");
	  }
	  /* update the counters */
	  haveRead=haveRead+actuallyRead;
	}
	extra=new String(block,0,0,n);
	/* in any case we want to add the thing we just read to what we had */
	/* read and then recursively call this code to finish up the read to the next */
	/* CR/LF. note: this leaves the {n} in the string so a parser can find it later */
	finalString=finalString+extra+readString();
      }
    }
    /* echo the protocol to the terminal */
    if (debugProto==true) {
      System.out.println("IMAP: <-- " + finalString);
    }
    /* send it home */
    return finalString;
  }
  /* This sends a command to the server and waits for the response */
  /* it accumulates the responses into the results Vector and returns */
  /* a status value in the boolean */
  public boolean roundTrip(String s, Vector result) throws InternalErr, IMAPException {
    /* initially, we are not done */
    boolean done = false;
    /* use this to do separation of words */
    StringTokenizer tokenizer;
    String token;
    /* overall result we see in the protocol */
    String resultToken=new String("Bogus!");
    /* write the string plush archive it */
    writeStringInternal(s,true);
    /* repeat until we get the message we are looking for */
    while (done==false) {
      /* read from the network  */
      try {
	s=readString();
      } catch (IOException e) {
	throw new IMAPException("IMAP Server IO Error: " + e.getMessage());
      }
      /* add to the end */
      result.addElement(s);
      /* rip out the first word an put it in token */
      tokenizer=new StringTokenizer(s);
      token=tokenizer.nextToken();
      /* are we done?*/
      if (token.equals("*")==false) {
	/* we should be able to find it in the hashtable */
	if (pendingRequests.containsKey(token)==true) {
	  /* found it ... get it out of the table */
	  pendingRequests.remove(token);
	  /* we are getting out */
	  done=true;
	  /* get the result of the command */
	  resultToken=tokenizer.nextToken();
	} else {
	  throw new InternalErr("Found a response to an IMAP roundtrip that I wasn't expecting!");
	}
      }
    }
    /* did we succeed or fail? */
    if (resultToken.equals("OK")==false) {
      /* bad protocol or failed command */
      return false;
    } else {
      /* we are cool */
      return true;
    }
  }
}

