/*
 *  $Id: Driver.java,v 1.6 1996/01/19 02:05:24 iansmith Exp $
 *  $Author: iansmith $
 *  $Log: Driver.java,v $
 * Revision 1.6  1996/01/19  02:05:24  iansmith
 * Documentation improvements in the javadoc comments.
 *
 * Added some code so the message about getting new mail now appears at the
 * time you see the prompt. This is using an AsynchronousAction object.
 *
 * If you type a blank line at the command prompt I now issue a NOOP
 * command so I get a check for new mail. This was a subtle thing: Since
 * I was doing caching on headers and bodies, once you read everything
 * you had no way to ever generate more protocol to see you had new mail
 * (via the RECENT message). This fixes that.
 *
 * The code now sets up its own Asynch.Action object to be the action
 * for BYE, RECENT, and EXPUNGE.
 *
 * Moved a bunch of code around so I could add the new commands: CHECK,
 * COPY, MOVE, EXPUNGE, CLOSE. This required some big changes to the way
 * the mainloop worked. I branched the code that does MOVE/COPY into its
 * own function and move and copy commands just use that. The CHECK and
 * CLOSE were easy so I just added them in place. The delete command
 * got modified so I now use the code that determines if a message
 * number is valid (a new function) to get the message number to delete
 * (move and copy use this too).
 *
 * Revision 1.5  1996/01/15  18:58:52  iansmith
 * Ripped out all the code which displays various things on the terminal and
 * moved it over to the Display class.
 *
 * Modified the output side of the MIME->HTML code to correctly deal with
 * HTML attachments (ala the ones that Netscape 2 produces!).
 *
 * Added the ? command to the main loop.
 *
 * Added the delete command to the main loop.
 *
 * Separated out the SELECT and EXAMINE commands.
 *
 * Modified the main loop so upper and lower case commands work the same.
 *
 * Revision 1.4  1996/01/10  22:07:06  iansmith
 * Just about totally rewrote this file this time. I added support for:
 * a command loop
 * a select command
 * a quit command
 * multiple IMAP servers (a prompt for this)
 * printing out the mail headers as a list
 * printing out the mail messages
 * printing out the MIME structure
 *
 *  Revision 1.3  1996/01/08  05:54:04  iansmith
 *  Added a throws InternalErr to all the API functions as some lower level
 *  code now throws it and I needed to pass it through.
 *
 *  Added a function to print a status line about an email message.
 *
 *  Added a function to display a mail message's headers.
 *
 *  Added a function to display a list of email addresses from a vector of
 *  MailAddresses.
 *
 *  Made some mods to the main program to accomodate use of the above.
 *
 *  Revision 1.2  1996/01/06 19:50:22  iansmith
 *  Added some RCS header info.
 *
 */

package GIT.GVU.IMAP.Tests;
import GIT.GVU.IMAP.*;
import java.lang.*;
import java.util.*;
import java.io.*;
/**
 * Class to implement a little bitty text based mailer.
 * 
 * @version $Id: Driver.java,v 1.6 1996/01/19 02:05:24 iansmith Exp $
 * @author Ian Smith
 */

public class Driver {
  /*
   * Print out a list of mailboxes in a nicely formatted way. This shows how
   * use the MailboxName object to figure out which are directories and which
   * are files.
   * 
   * @param Connection object which has had the LIST call made on it
   * @param s boolean which controls whether to print selectable mailboxes (true) or unselectable mailboxes (false)
   *
   */
  static void printMailboxes(Connection c, boolean s) {
    Vector v=c.getMailboxNames();
    MailboxName mn;
    int i;
    String name;

    /* loop over all names */
    for (i=0;i<v.size();++i) {
      /* extract mailbox name */
      mn=(MailboxName)v.elementAt(i);
      /* check for selectability equal to our param */
      if (mn.getSelectable()==s) {
	/* got it! */
	name=mn.getName();
	/* format this name */
	name=Util.makeWidth(name,40);
	/* if its interesting, add that to the name */
	if (mn.getInteresting()==true) {
	  name=name+ "(Interesting)";
	}
	/* print it! */
	System.out.println(name);
      }
    }
  }
  /**
   * Function to demo how to list the mailboxes.
   *
   * @param c connection on which to perform the LIST command
   * @exception InternalErr is propagated
   * @exception IMAPException is propagated
   */
  public static void listMailboxes(Connection c) throws InternalErr, IMAPException {
    /* send the LIST command and gather results */
    c.LIST("~iansmith/Mail/","%");
    /* print out the selectable ones */
    System.out.println("Selectable: (files):");
    printMailboxes(c,true);
    /* print out the not selectable ones */
    System.out.println("Not Selectable: (directories):");
    printMailboxes(c,false);
  }
  /**
   * Function to demo how to use CREATE and RENAME. Note: This is currently dead code.
   * 
   * @param c connection to perform the transactions on
   * @exception InternalErr is propagated
   * @exception IMAPException is propagated
   */
  public static void createThenRenameThenDelete(Connection c) throws InternalErr , IMAPException{
    if (c.CREATE("~iansmith/Mail/foobar")==false) {
      System.out.println("Failed to create mailbox");
    }
    if (c.RENAME("~iansmith/Mail/foobar","~iansmith/Mail/fripfrap")==false) {
      System.out.println("Failed to rename mailbox");
    }
    /* should fail,its not there! */
    if (c.DELETE("~iansmith/Mail/foobar")==false) {
      System.out.println("Failed to DELETE foobar");
    }
    /* this should work */
    if (c.DELETE("~iansmith/Mail/fripfrap")==false) {
      System.out.println("Failed to DELETE fripfrap");
    }
  }
  /*
   * Utility function to write a sting to a fileoutputstream.
   *
   * @param FileOutputStream fs the stream to write to
   * @param String s the string to write
   * @exception java.io.IOException is propagated
   */
  static void putStringOnStream(FileOutputStream fs, String s) throws java.io.IOException {
    byte data[]=new byte[s.length()];
    /* conver to an array of bytes and then write it */
    s.getBytes(0,s.length(),data,0);
    fs.write(data,0,s.length());
  }
  /**
    * This function writes out a MIME message as HTML to a file. 
    * 
    * 
    * @param BodyPart bp First BodyPart of the message
    * @param FileOutputStream fs The file to write the HTML on
    * @param String name is the base name of the message for generating tmp files
    * @param int num number for generating external files
    * @exception java.io.IOException is propagated
    * @exception InternalErr is propagated
    * @exception IMAPException is propagated
     * @return boolean true if we succeeded with the whole message
   *
   */
  public static boolean writeMIME(BodyPart bp, FileOutputStream fs, String name, int num) throws java.io.IOException, InternalErr, IMAPException {
    BodyPart child;
    byte data;
    String tmp,fname,binaryData;
    FileOutputStream outerFile;

    /* is this a multipart part? */
    if (bp.getFirstChild()!=null) {
      /* it is, recurse down children */
      child=bp.getFirstChild();
      /* do all children */
      while (child!=null) {
	/* test each child */
	if (writeMIME(child,fs,name,num)==false) {
	  return false;
	}
	/* update the number in case they wrote on it */
	++num;
	/* next*/
	child=child.getNextSibling();
      }
      /* all children worked */
      return true;
    }
    System.out.println("Writing out part " + bp.getPartNumber() + "...");
    /* is this a text message ? */
    if (bp.getMIMEType().equalsIgnoreCase("TEXT")==true) {
      if (bp.getMIMESubType().equalsIgnoreCase("PLAIN")==true) {
	/* preformatted data */
	Driver.putStringOnStream(fs,"<PRE>\n");
	bp.FETCH();
	Driver.putStringOnStream(fs,bp.getData());
	Driver.putStringOnStream(fs,"</PRE><HR>\n");
	return true;
      }
      /* HTML? */
      if (bp.getMIMESubType().equalsIgnoreCase("HTML")==true) {
	/* just get the data an echo it */
	bp.FETCH();
	Driver.putStringOnStream(fs,bp.getData());
	Driver.putStringOnStream(fs,"<HR>");
	return true;
      }
    }
    /* we know about IMAGE/GIF  and IMAGE/JPEG */
    /* we want to inline those */
    if (bp.getMIMEType().equalsIgnoreCase("IMAGE")==true) {
      if ((bp.getMIMESubType().equalsIgnoreCase("GIF")==true) ||
          (bp.getMIMESubType().equalsIgnoreCase("JPEG")==true)) {
	/* now make sure its base64 encoded */
	if (bp.getEncoding().equalsIgnoreCase("BASE64")==true) {
	  /* ok its a type we know */
	  Driver.putStringOnStream(fs,"<P>Embedded Image:<P><IMG SRC=");
	  /* construct last part of tmp file */
	  fname="."+(new Integer(num)).toString() + "." + bp.getMIMESubType().toLowerCase();
	  /* try to open the file */
	  try {
	    outerFile=new FileOutputStream(name+fname);
	    /* put out the HTML */
	    Driver.putStringOnStream(fs,name+fname+">\n<HR>");
	    /* decode the data */
	    bp.FETCH();
	    binaryData=MIME.decodeBase64(bp.getData());
	    /* output the data */
	    Driver.putStringOnStream(outerFile,binaryData);
	    /* close it up */
	    outerFile.close();
	    /* done with HTML now do file output */
	  } catch (IOException e) {
	    /* couldn't open it */
	    return false;
	  }
	  /* it worked */
	  return true;
	}
      }
    }
    /* now we've tried the types we know so its some type we don't understand */
    /* construct last part of tmp file by looking at the MIME typing*/
    fname="."+(new Integer(num)).toString() + 
          MIME.findExtension(bp.getMIMEType()+"/"+bp.getMIMESubType());
    /* try to open the file */
    try {
      outerFile=new FileOutputStream(name+fname);
      /* prefix it */
      Driver.putStringOnStream(fs,"Can't inline decode <A HREF="+name+fname+"> "
                                  + "  a part of this message of type" + 
                                  bp.getMIMEType()+"/"+bp.getMIMESubType() +
                                  "</A><P><HR>");
      /* make sure its in cache */
      bp.FETCH();
      /* test for base64 */
      if (bp.getEncoding().equalsIgnoreCase("BASE64")==true) {
	binaryData=MIME.decodeBase64(bp.getData());
      } else {
	/* bogus! we assume we don't need decoding just because its not base 64 */
	binaryData=bp.getData();
      }
      /* put out the data we don't know about */
      Driver.putStringOnStream(outerFile,binaryData);
    } catch (IOException e) {
      /* can't get it to work ... crap out*/
      return false;
    }
    /* we succeeded, tell the caller ********/
    return true;
  }
  /** 
   * This is the system's main loop... read a string process the command then repeat
   * @param Connection c the connection we are using
   * @param Protocol p the prtocol which is the owner of this connection
   * @exception InternalErr is propagated
   * @exception IMAPException is propagated
   * @exception java.io.IOException is propagated
   */
  public static void mainloop(Connection c, Protocol protocol) throws InternalErr, IMAPException, java.io.IOException {
    boolean done=false,gotnumber;
    String command,mboxname,first,num;
    StringTokenizer st;
    int n,lastRead=0,exist;
    Message message;

    /* loop until we get a quit */
    while (done==false) {
      /* before we print a prompt, do we need to print a new mail notice */
      if (Driver.newMailCount!=0) {
	/* the notifier set it, tell user */
	System.out.println("You have new messages in the selected mailbox (" + 
                            Driver.newMailCount + ")");
	/* reset the counter */
	Driver.newMailCount=0;
      }
      /* print prompt */
      System.out.print("HackMail 0.9 ");
      /* do we have a mailbox? */
      if (c.getMailbox()!=null) {
	/* have they read any yet? */
	if (lastRead!=0) {
	  /* yep, tell them which */
	  System.out.print("("+lastRead +" of ");
	} else {
	  /* no */
	  System.out.print("(");
	}
	/* how many messages in this box? */
	exist=c.getMailbox().getExists();
	if (exist==0) {
	  System.out.print("no messages) ");
	}
	if (exist==1) {
	  System.out.print("1 message) ");
	}
	if (exist>1) {
	  System.out.print(exist + " messages) ");
	}
      }
      /* print end of prompt */
      System.out.print(" ? for help> ");
      System.out.flush();
      /* read the command */
      command=Util.getLine();
      /* if empty string is input, try sending a noop */
      if (command.equals("")==true) {
	/* this is effectively a check for new mail */
	if (protocol.NOOP()==false) {
	  System.out.println("Serious problem: Failed a NOOP() command.");
	}
	/* helpful message */
	System.out.println("? for help");
	continue;
      }
      /* get a tokenizer */
      st=new StringTokenizer(command);
      /* try convert to a number */
      try {
	n=Integer.parseInt(command,10);
	gotnumber=true;
      } catch (NumberFormatException e) {
	n=0;
	gotnumber=false;
      }
      /* if we got a number, lets display it */
      if (gotnumber==true) {
	/* do we have a mailbox? */
	if (c.getMailbox()==null) {
	  System.out.println("No mailbox selected yet!");
	  continue;
	}
	/* got a mailbox... lets check to see how many messages and check */
	/* message number validity ...*/
	if ((n<1) || (n>c.getMailbox().getExists())) {
	  System.out.println("Invalid message specified! Message numbers are 1-" + 
                             c.getMailbox().getExists());
	  continue;
	}
	/* valid number and box, print it out */
	Display.displayMessage(c,n);
	/* update the last message read */
	lastRead=n;
	/* read next command... */
	continue;
      }
      /* get first thing on the command */
      first=st.nextToken();
      first=first.toLowerCase();

      /* didn't get a number... lets look for select */
      if ((first.equals("s")==true) || (first.equals("select")==true) ||
          (first.equals("e")==true) || (first.equals("examine")==true))  {
	/* do we need to prompt them */
	if (st.hasMoreTokens()==false) {
	  /* ask what mailbox they want */
	  System.out.print("Select/Examine what mailbox?");
	  System.out.flush();
	  mboxname=Util.getLine();
	} else {
	  /* they supplied it on the command line */
	  mboxname=st.nextToken();
	}
	/* is it select or examine ? */
	if (first.substring(0,1).equals("s")==true) {
	  /* try to select it */
	  if (c.SELECT(mboxname)==false) {
	    System.out.println("Unable to SELECT mailbox " + mboxname);
	    continue;
	  } else {
	    /* got it */
	    System.out.println("Warning: Using SELECT! Changes will be permanent!");
	  }
	} else {
	  /* must be examine */
	  if (c.EXAMINE(mboxname)==false) {
	    System.out.println("Unable to EXAMINE mailbox " + mboxname);
	    continue;
	  } else {
	    /* got it */
	    System.out.println("Warning: Using EXAMINE! Changes will not be permanent!");
	  }
	}
	/* install our notifier for both expunge and recent */
	c.getMailbox().setExpungeAction(notifier);
	/* remember: We won't get a call for new mail on the fact that we just */
	/* opened the mailbox */
	c.getMailbox().setRecentAction(notifier);
	/* show the user the info about this box */
	Display.printMailboxInfo(c.getMailbox(),false);
	/* update the number of last message read (now they haven't read anything)*/
	lastRead=0;
	/* show them the headers, just for grins */
	Display.displayHeaders(c.getMailbox());
	/* read next command */
	continue;
      }
      /* lets look for a quit command */
      if ((first.equals("q")==true) || (first.equals("quit")==true)) {
	done=true;
      }
      /* try for headers*/
      if ((first.equals("h")==true) || (first.equals("headers")==true)) {
	/* make sure we have a mailbox */
	if (c.getMailbox()==null) {
	  System.out.println("Can't print headers with no mailbox selected!");
	} else {
	  /* show the headers */
	  Display.displayHeaders(c.getMailbox());
	}
	/* read next command */
	continue;
      }
      /* try for a delete */
      if ((first.equals("d")==true) || (first.equals("delete")==true)) {
	/* what message ?*/
	n=readMessageNumber("DELETE",c,st);
	/* is it bogus number ? */
	if (n==-1) {
	  /* bogus number, read next command */
	  continue;
	}
	/* just delete the thing */
	deleteMessage(n,c);
	/* in any case, lets get the next command */
	continue;
      }
      /* ok, lets look for a question mark */
      if (first.equals("?")==true) {
	/* print a help message */
	System.out.println("Commands:");
	System.out.println("\ts\tSelect [mailbox name]");
	System.out.println("\te\tExamine [mailbox name] (select read-only)");
	System.out.println("\th\tHeaders");
	System.out.println("\td\tDelete [message number]");
	System.out.println("\tk\tcheckpoint");
	System.out.println("\tt\tcreaTe [mailbox name]");
	System.out.println("\tc\tCopy [message number] [mailbox name]");
	System.out.println("\tm\tMove [message number] [mailbox name] (copy+delete)");
	System.out.println("\tx\teXpunge (Actually performs deletes)");
	System.out.println("\tl\tcLose (Unselect this mailbox))");
	System.out.println("or just type a message number to read it");
      }
      /* ok, lets look for a checkpoint command */
      if ((first.equals("k")==true) || (first.equals("checkpoint")==true)) {
	/* make sure there is a mailbox */
	if (c.getMailbox()==null) {
	  System.out.println("You must select or examine a mailbox to CHECKpoint it");
	  continue;
	}
	/* ok, they have a mailbox, try to checkpoint it */
	if (c.getMailbox().CHECK()==true) {
	  System.out.println("Checkpoint succeeded");
	} else {
	  System.out.println("Checkpoing failed");
	}
	/* in case, go read the next command */
	continue;
      }
      /* lets look for a close command */
      if ((first.equals("l")==true) || (first.equals("close")==true)) {
	/* make sure there is a mailbox */
	if (c.getMailbox()==null) {
	  System.out.println("You must select or examine a mailbox to CLOSE it");
	  continue;
	}
	/* got a box, so lets close */
	if (c.getMailbox().CLOSE()==true) {
	  System.out.println("Mailbox closed");
	} else {
	  System.out.println("Failed to close mailbox: CLOSE failed");
	}
	/* go read next command */
	continue;
      }
      /* how about an EXPUNGE */
      if ((first.equals("x")==true) || (first.equals("expunge")==true)) {
	/* make sure there is a mailbox */
	if (c.getMailbox()==null) {
	  System.out.println("You must select or examine a mailbox to EXPUNGE it");
	  continue;
	}
	/* expunge it */
	if (c.getMailbox().EXPUNGE()==true) {
	  /* expunge succeeded, relist the headers */
	  Display.displayHeaders(c.getMailbox());
	  /* reset the message number */
	  lastRead=0;
	} else {
	  System.out.println("EXPUNGE failed");
	}
	/* in any case, read next command */
	continue;
      }
      /* might be a create command */
      if ((first.equals("t")==true) || (first.equals("create")==true)) {
	/* all we need is a connection, so go to it ... see if they typed it on command line*/
	if (st.hasMoreTokens()==true) {
	  mboxname=st.nextToken();
	} else {
	  /* read from terminal */
	  System.out.print("Create what mailbox? ");
	  mboxname=Util.getLine();
	}
	/* try it */
	if (c.CREATE(mboxname)==true) {
	  System.out.println("Created mailbox " + mboxname+ " successfully");
	} else {
	  System.out.println("Unable to create mailbox " + mboxname);
	}
	/* read next command */
	continue;
      }
      /* look for a copy command */
      if ((first.equals("c")==true) || (first.equals("copy")==true)) {
	/* go execute a copy command */
	copyMessage(st,c,false,"COPY");
	/* get next command */
	continue;
      }
      /* look for a move command */
      if ((first.equals("m")==true) || (first.equals("move")==true)) {
	/* go execute a copy command */
	copyMessage(st,c,true,"MOVE");
	/* get next command */
	continue;
      }
    }
  }
  /**
   * This function deletes a message. It expects a good message number and a connection
   * with a mailbox attached to it .
   * 
   * @param int n message number to delete
   * @param Connection c connection object to manipulate
   * @exception InternalErr is propagated
   * @exception IMAPException is propagated
   */
  static void deleteMessage(int n, Connection c) throws InternalErr, IMAPException {
    Message message;
    /* ok, it worked, lets do it ...grab the message */
    c.getMailbox().FETCH(n);
    message=c.getMailbox().getMessage(n);
    /* did it work? */
    if (message==null) {
      System.out.println("STRANGE THING: Bad message value in delete");
    } else {
      /* cool... send the STORE message */
      if (message.STORE("+","\\Deleted")==true) {
	System.out.println("Message " + n + " marked for deletion");
      } else {
	/* failed to delete */
	System.out.println("Failed trying to delete message " + n);
      }
    }
    /* go home */
    return;
  }
  /** 
   * This function tries to copy a message to a mailbox 
   * 
   * @param StringTokenizer st tokenizer with the input tokens (or not) on it 
   * @param Connection c connection object we are working with  
   * @param Boolean d delete the message after copying it 
   * @param String cmd command we are executing (usually COPY or MOVE)
   * @exception java.io.IOException is propagated
   * @exception InternalErr is propagated
   * @exception IMAPException is propagated
   */
  static void copyMessage(StringTokenizer st, Connection c, boolean d, String cmd) throws java.io.IOException, InternalErr, IMAPException {
    int n;
    String mboxname;
    Boolean bool;

    /* got a mailbox get a number */
    n=readMessageNumber(cmd,c,st);
    /* bad number? */
    if (n==-1) {
      /* n is bad number, so bail out */
      return;
    }
    /* look for another token which is the mailbox name */
    if (st.hasMoreTokens()==true) {
      mboxname=st.nextToken();
    } else {
      /* no more tokens, read from command line */
      System.out.print(cmd + " to what mailbox? (default prefix "+
                       defaultPrefix + " )? ");
      mboxname=Util.getLine();
    }
    /* excute copy command */
    bool=c.getMailbox().COPY(n,mboxname);
    /* was it a null which means try again ? */
    if (bool==null) {
      /* try using the default */
      bool=c.getMailbox().COPY(n,defaultPrefix+mboxname);
      /* check for problem with *2nd* move/copy */
      if (bool==null) {
	System.out.println(cmd + " failed. Try using CREATE to create the mailbox "
                             + mboxname );
	return;
      }
    }
    /* now we know its true or false*/
    /* NOTE: WE USE EQUALS HERE NOT == !! */
    if (bool.equals((new Boolean(true)))) {
      System.out.println(cmd + " succeeded on Message " + 
                         (new Integer(n)).toString());
      /* now if we are doing a move, just try to delete now */
      if (d==true) {
	deleteMessage(n,c);
      }
    } else {
      System.out.println("Unable to " + cmd + "  message" + 
                         (new Integer(n)).toString());
    }
    /* we're done */
    return;
  }

  /**
   * Utility function to read a message number. It checks for selected mailbox or not.
   * 
   * @param String cmd command we are executing (for error messages)
   * @param Connection c connection object to use
   * @exception java.io.IOException is propagated
   * @exception InternalErr is propagated
   * @return message number which they entered or -1 on error
   */
  static int readMessageNumber(String cmd, Connection c, StringTokenizer st) throws java.io.IOException, InternalErr{
    String num;
    int n=-1;
    boolean gotnumber;
    
    /* make sure there is a mailbox */
    if (c.getMailbox()==null) {
      System.out.println("Unable to perform " + cmd + " unless you have selected or examined a mailbox");
      return -1;
    }
    /* ok, what number? */
    if (st.hasMoreTokens()==true) {
      num=st.nextToken();
    } else {
      System.out.print(cmd + " what message ? ");
      num=Util.getLine();
    }
    /* try convert to a number */
    try {
      n=Integer.parseInt(num,10);
      gotnumber=true;
    } catch (NumberFormatException e) {
      gotnumber=false;
    }
    /* did it work ? */
    if (gotnumber==false) {
      System.out.println("Unable to find parse " + num + " as a number");
      return -1;
    }
    /* is it a valid message number? */
    if ((n<1) || (n>c.getMailbox().getExists())) {
      System.out.println("Invalid message specified! Message numbers are 1-" + 
                          c.getMailbox().getExists());
      return -1;
    }
    /* everything's cool */
    return n;
  }
  /**
   * Perform the login procedure
   * 
   * @exception InternalErr is propagated
   * @exception IMAPException is propagated
   * @return a connection object or null if the login failed
   */
  public static Connection login(Protocol p) throws InternalErr, IMAPException, java.io.IOException {
    Connection c;
    String username, password, servername;
    int i;

    /* create a socket to the server */
    System.out.print("IMAP Server : ");
    servername=Util.getLine();
    p.connect(servername);
    /* ask the user the questions */
    System.out.print("Username    : ");
    username=Util.getLine();
    System.out.print("Password    : ");
    System.out.flush();
    /* XXX ECHOES PASSWORD TO SCREEN XXX */
    password=Util.getLine();
    /* need to clear the screen and we really can't do this in a machine */
    /* independant way ... so just output 100 lines*/
    for (i=0; i<100; ++i) {
      System.out.println("");
    }
    /* try to log in */
    if (p.LOGIN(username,password)==false) {
      System.out.println("LOGIN failed!!");
      /* try to logout */
      if (p.LOGOUT()==false) {
	/* wow, that didn't even work */
	System.out.println("The world is seriously hosed up!");
      }
      return null;
    }
    /* logged in ok */
    /* again, we need to clear the screen so we just dump out some lines */
    for (i=0; i<100; ++i) {
      System.out.println("");
    }
    return p.getConnection();
  }

  /* main program entry point */
  public static void main(String argv[]) {
    /* catch IMAP and Internal Errors below */
    /* create a protocol object */
    Connection c;
    Protocol p;
    
    /* create a protcol object */
    p=new Protocol();
    try {
      /* run login procedure */
      c=login(p);
      /* did we log in ok? */
      if (c!=null) {
	/* before we get too far, lets install the notifier as the callback object */
	/* for BYE */
	p.setByeAction(notifier);
	/* execute the system mainloop */
	mainloop(c,p);
	if (p.LOGOUT()==false) {
	  System.out.println("Weird! Unable to logout!");
	}
      }
      /* catch any errors that occurred here */
    } catch (IMAPException e) {
      /* you can use e.getMessage() to get the value of e as a string */
      /* you can also use e.printStackTrace() to print the current value */
      /* of the stack for debugging */
      /* just print out the value of the exception */
      System.out.println("IMAP Exception! " + e);
    } catch (InternalErr e) {
      /* again, just print out its value */
      System.out.println("Internal Error! " + e);
    } catch (java.io.IOException e) {
      System.out.println("IOException: Likely caused by terminal input problem: " + e);
    }
  }
  /** 
   * This holds our Notifier object. 
   */
  static Notify notifier;
  /**
   * This variable gets set by the Notifier to tell us there is new mail.
   * If it is non-zero then we need to tell the user there is new mail. 
   */
  static int newMailCount;
  /**
   * This variable holds the default email folder prefix for the mailer.
   */
    static final String defaultPrefix="~/HM/";
  /**
   * This is the  function that gets run when our object gets loaded.
   */
  static {
    /* initialize our notifier object */
    notifier=new Notify();
    /* when we start there is no new mail */
    newMailCount=0;
  }
}
